Friend Class MachineDefinition
    Inherits MachineDefinitionBase
    Implements IMachineDefinition_2_0
	
	''' <summary>
	''' The version of this machine definition as shown by choosing the Version command from the Help menu of the Configure Machine Definition dialog
	''' </summary>
	Private version As System.Version =  New System.Version("1.0.1")
	
    ''' <summary>
    ''' Local variable to store the format for the number of decimal places after the point.
    ''' </summary>
    Private format As String
	
	''' <summary>
	''' Local variable that is used to store the configuration.
	''' </summary>
	''' <remarks>Any changes made by the user from the 'Configure Machine Definition' command are assigned to this variable.</remarks>
	Private _configuration As Configuration
	
    ''' <summary>
    ''' Local variable to store the maximum feedrate.
    ''' </summary>
	Private maxFeedrate As String
	
    ''' <summary>
    ''' Local variable to store the current cutter radius compensation.
    ''' </summary>
	Private currentCRC As String
	
	''' <summary>
	''' Largest value that Zero can be represented by, with respect to epsilon based comparison
	''' </summary>
	Private Const ZERO_MAX As Double = 1.0e-6
	
	''' <summary>
	''' Local variables to store the previous positions
	''' </summary>
	Private previousX, previousY, previousZ, previousCCX, previousCCY As Double
	
	''' <summary>
	''' Local variable to store if the first Circle Centre has been output
	''' </summary>
	Private firstCC As Boolean
	
	''' <summary>
	''' Local variable to store a collection of tool offsets
	''' </summary>
	Private toolOffsets As System.Collections.Generic.List(Of Double)
	
	''' <summary>
	''' Local variable to store a collection of tool numbers
	''' </summary>
	Private toolNumbers As System.Collections.Generic.List(Of Double)
		
	''' <summary>
	''' Local variable to store a collection of tool radiuses
	''' </summary>
	Private toolRadiuses As System.Collections.Generic.List(Of Double)
	
#Region "Events"

	''' <summary>
	''' This event handler is invoked when this Machine Definition has finished compiling.
	''' </summary>
	''' <param name="e">Provides data for the compiled event.</param>
    ''' <remarks>Here is where you should perform any one time initialization.</remarks>
	Private Sub Complied(ByVal e As CompiledEventArgs) Implements IMachineDefinition_2_0.Compiled
						
	End Sub

    ''' <summary>
    ''' This event handler is invoked first and is where you configure the general settings.
    ''' </summary>
    ''' <param name="e">Provides data for the general settings event.</param>
    Private Sub GeneralSettings(ByVal e As GeneralSettingsEventArgs) Implements IMachineDefinition_2_0.GeneralSettings

        e.OutputLineNumbers = True
        e.StartLineNumber = 1
        e.LineNumberIncrement = 1
        e.ArcBreak = ArcBreak.BreakAt180Degrees
		e.NcCodeExtension = "h"
		e.SubRoutinesAtStart = False
		e.LineNumbering = LineNumbering.Sequentially
		e.SubRoutinesStartNumber = 1
		e.SubRoutinesNumberIncrement = 1
		
        'Metric parts need 3 decimal places after the point, imperial 4
        If e.Metric Then
            format = "0.0##"
            e.MinimumArcLength = 0.005
            e.MinimumArcRadius = 0.005
            e.MaximumArcRadius = 10000
        Else
            format = "0.0###"
            e.MinimumArcLength = 0.0002
            e.MinimumArcRadius = 0.0002
            e.MaximumArcRadius = 400
		End If

		maxFeedrate = "F" & _configuration.MaxFeedrate
		
		toolOffsets = New System.Collections.Generic.List(Of Double)
		toolNumbers = New System.Collections.Generic.List(Of Double)
		toolRadiuses = New System.Collections.Generic.List(Of Double)
		
    End Sub

	''' <summary>
	''' This event handler is invoked so the start of the file can be created.
	''' </summary>
	''' <param name="e">Provides data for the program start event.</param>
	''' <remarks>Here is where you should create the begining of the file, for example the % sign as used by a Fanuc.</remarks>
	Private Sub FileStart(ByVal e As FileStartEventArgs) Implements IMachineDefinition_2_0.FileStart

	End Sub


    ''' <summary>
    ''' This event handler is invoked so the start of the program can be created.
    ''' </summary>
    ''' <param name="e">Provides data for the program start event.</param>
    ''' <remarks>Here is where you should create any preparatory code.</remarks>
    Private Sub ProgramStart(ByVal e As ProgramStartEventArgs) Implements IMachineDefinition_2_0.ProgramStart

		For Each toolNumber As Double In e.ToolNumbers
	        NcCode.LineNumber()
	        NcCode.NewString("TOOL DEF " & toolNumber & " L0 R0")
	        NcCode.NewLine()
        Next
		currentCRC = "R0"

		NcCode.LineNumber()
		NcCode.NewString("STOP M5")
	    NcCode.NewLine()
		
    End Sub

    ''' <summary>
    ''' This event handler is invoked for every tool change.
    ''' </summary>
    ''' <param name="e">Provides data for the tool change event.</param>
    Private Sub ToolChange(ByVal e As ToolChangeEventArgs) Implements IMachineDefinition_2_0.ToolChange
				
        NcCode.LineNumber()
        NcCode.NewString("TOOL CALL " & e.ToolNumber & " Z S" & e.SpindleSpeed)
        NcCode.NewLine()
		
		toolOffsets.Add(e.ToolOffsetNumber)
    	toolNumbers.Add(e.ToolNumber)	
		toolRadiuses.Add(e.ToolDiameter / 2)
		
		firstCC = True
		
    End Sub

    ''' <summary>
    ''' This event handler is invoked before every tool change, unless it is the first, where it is not called.
    ''' </summary>
    ''' <param name="e">Provides data for the cancel tool change event.</param>
    Private Sub CancelToolChange(ByVal e As CancelToolChangeEventArgs) Implements IMachineDefinition_2_0.CancelToolChange
        
		NcCode.LineNumber()
		NcCode.NewString("STOP M5")
	    NcCode.NewLine()

    End Sub

    ''' <summary>
    ''' This event handler is invoked for the first position after a tool change for Profiling and Pocketing operations.
    ''' </summary>
    ''' <param name="e">Provides data for the first position after tool change event.</param>
    Private Sub FirstPositionAfterToolChange(ByVal e As FirstPositionAfterToolChangeEventArgs) Implements IMachineDefinition_2_0.FirstPositionAfterToolChange

        NcCode.LineNumber()
        NcCode.NewString("L ")
        NcCode.NewDimensionedWord("X", e.PositionX.ToString(format, CultureInfo), Modal.Off)
        NcCode.NewDimensionedWord("Y", e.PositionY.ToString(format, CultureInfo), Modal.Off)
        NcCode.NewString("R0 " & maxFeedrate & " " & e.ToolDirection)
        NcCode.NewLine()
        NcCode.LineNumber()
		NcCode.NewString("L ")
        NcCode.NewDimensionedWord("Z", e.PositionZ.ToString(format, CultureInfo), Modal.Off)
        NcCode.NewString("R0 " & maxFeedrate & " " & e.Coolant)
        NcCode.NewLine()
		previousX = e.PositionX
		previousY = e.PositionY
		previousZ = e.PositionZ
		
    End Sub

    ''' <summary>
    ''' This event handler is invoked for all Profiling and Pocketing rapid positions, other than
    ''' the <see cref="FirstPositionAfterToolChange"/>.
    ''' </summary>
    ''' <param name="e">Provides data for the rapid position event.</param>
    Private Sub RapidPositiong(ByVal e As RapidPositionEventArgs) Implements IMachineDefinition_2_0.RapidPositioning

		Dim moveType As MoveType = MovementType(previousX, previousY, previousZ, e.AbsX, e.AbsY, e.AbsZ)
		If moveType = MoveType.XY Then
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("X", e.AbsX.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewDimensionedWord("Y", e.AbsY.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString("R0 " & maxFeedrate & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.Z Then
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("Z", e.AbsZ.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString("R0 " & maxFeedrate & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.XYZ Then
			MessageBox.Show("Three axis rapid move detected", "Error", MessageBoxButtons.OK, MessageBoxIcon.Information)
			Throw New System.Exception("Nc code generation has been terminated")
		End If
		previousX = e.AbsX
		previousY = e.AbsY
		previousZ = e.AbsZ

    End Sub

    ''' <summary>
    ''' This event handler is invoked for every linear interpolation move.
    ''' </summary>
    ''' <param name="e">Provides data for the linear interpolation event.</param>
    Private Sub LinearInterpolation(ByVal e As LinearInterpEventArgs) Implements IMachineDefinition_2_0.LinearInterpolation

		Dim moveType As MoveType = MovementType(previousX, previousY, previousZ, e.AbsEndX, e.AbsEndY, e.AbsEndZ)
		If moveType = MoveType.XY Then
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("X", e.AbsX.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewDimensionedWord("Y", e.AbsY.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString(currentCRC & " F" & e.FeedRate.ToString(0, CultureInfo) & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.Z Then
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("Z", e.AbsZ.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString(currentCRC & " F" & e.FeedRate.ToString(0, CultureInfo) & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.XYZ Then
			MessageBox.Show("Three axis linear move detected", "Error", MessageBoxButtons.OK, MessageBoxIcon.Information)
			Throw New System.Exception("Nc code generation has been terminated")
		End If
		previousX = e.AbsX
		previousY = e.AbsY
		previousZ = e.AbsZ

    End Sub

    ''' <summary>
    ''' This event handler is invoked when Cutter Radius Compensation (CRC) is applied on the left.
    ''' </summary>
    ''' <param name="e">Provides data for the apply CRC left event</param>
    ''' <remarks>This is a linear move. CRC is not applied on circular interpolation moves</remarks>
    Private Sub ApplyCRCLeft(ByVal e As ApplyCRCLeftEventArgs) Implements IMachineDefinition_2_0.ApplyCRCLeft

        currentCRC = "RL"
		Dim moveType As MoveType = MovementType(previousX, previousY, previousZ, e.AbsEndX, e.AbsEndY, e.AbsEndZ)
		If moveType = MoveType.XY Then
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("X", e.AbsX.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewDimensionedWord("Y", e.AbsY.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString(currentCRC & " F" & e.FeedRate.ToString(0, CultureInfo) & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.Z Then
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("Z", e.AbsZ.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString(currentCRC & " F" & e.FeedRate.ToString(0, CultureInfo) & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.XYZ Then
			MessageBox.Show("Three axis linear move detected", "Error", MessageBoxButtons.OK, MessageBoxIcon.Information)
			Throw New System.Exception("Nc code generation has been terminated")
		End If
		previousX = e.AbsEndX
		previousY = e.AbsEndY
		previousZ = e.AbsEndZ

    End Sub

    ''' <summary>
    ''' This event handler is invoked when Cutter Radius Compensation (CRC) is applied on the right.
    ''' </summary>
    ''' <param name="e">Provides data for the apply CRC right event</param>
    ''' <remarks>This is a linear move. CRC is not applied on circular interpolation moves</remarks>
    Private Sub ApplyCRCRight(ByVal e As ApplyCRCRightEventArgs) Implements IMachineDefinition_2_0.ApplyCRCRight

		currentCRC = "RR"
		Dim moveType As MoveType = MovementType(previousX, previousY, previousZ, e.AbsEndX, e.AbsEndY, e.AbsEndZ)
		If moveType = MoveType.XY Then
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("X", e.AbsX.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewDimensionedWord("Y", e.AbsY.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString(currentCRC & " F" & e.FeedRate.ToString(0, CultureInfo) & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.Z Then
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("Z", e.AbsZ.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString(currentCRC & " F" & e.FeedRate.ToString(0, CultureInfo) & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.XYZ Then
			MessageBox.Show("Three axis linear move detected", "Error", MessageBoxButtons.OK, MessageBoxIcon.Information)
			Throw New System.Exception("Nc code generation has been terminated")
		End If
		previousX = e.AbsEndX
		previousY = e.AbsEndY
		previousZ = e.AbsEndZ

    End Sub

    ''' <summary>
    ''' This event handler is invoked when Cutter Radius Compensation (CRC) is canceled.
    ''' </summary>
    ''' <param name="e">Provides data for the cancel CRC right event.</param>
    ''' <remarks>This is a linear move. CRC is not cancelled on circular interpolation moves.</remarks>
    Private Sub CancelCRC(ByVal e As CancelCRCEventArgs) Implements IMachineDefinition_2_0.CancelCRC

        currentCRC = "R0"
		Dim moveType As MoveType = MovementType(previousX, previousY, previousZ, e.AbsX, e.AbsY, e.AbsZ)
		If moveType = MoveType.XY Then
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("X", e.AbsX.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewDimensionedWord("Y", e.AbsY.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString(currentCRC & " F" & e.FeedRate.ToString(0, CultureInfo) & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.Z Then
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("Z", e.AbsZ.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString(currentCRC & " F" & e.FeedRate.ToString(0, CultureInfo) & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.XYZ Then
			MessageBox.Show("Three axis linear move detected", "Error", MessageBoxButtons.OK, MessageBoxIcon.Information)
			Throw New System.Exception("Nc code generation has been terminated")
		End If
		previousX = e.AbsX
		previousY = e.AbsY
		previousZ = e.AbsZ

	End Sub

    ''' <summary>
    ''' This event handler is invoked for every counter clockwise circular interpolation move.
    ''' </summary>
    ''' <param name="e">Provides data for the clockwise circular interpolation event</param>
    Private Sub CWInterpolation(ByVal e As CWInterpolationEventArgs) Implements IMachineDefinition_2_0.CWInterpolation

   		Dim moveType As MoveType = MovementType(previousX, previousY, previousZ, e.AbsEndX, e.AbsEndY, e.AbsEndZ)
		If moveType = MoveType.XY Then
			If firstCC OrElse Not Equality(previousCCX, e.AbsCentreX) OrElse Not Equality(previousCCY, e.AbsCentreY)
				NcCode.LineNumber()
				NcCode.NewString("CC X" & e.AbsCentreX.ToString(format, CultureInfo) & " Y" & e.AbsCentreY.ToString(format, CultureInfo))
				NcCode.NewLine()
				firstCC = False
			End If
			previousCCX = e.AbsCentreX
			previousCCY = e.AbsCentreY
			NcCode.LineNumber()
        	NcCode.NewString("C ")
			NcCode.NewDimensionedWord("X", e.AbsEndX.ToString(format, CultureInfo), Modal.Off)
       		NcCode.NewDimensionedWord("Y", e.AbsEndY.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString("DR- ")
			NcCode.NewString(currentCRC & " F" & e.FeedRate.ToString(0, CultureInfo) & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.XYZ Then
			MessageBox.Show("Three axis CW circular move detected", "Error", MessageBoxButtons.OK, MessageBoxIcon.Information)
			Throw New System.Exception("Nc code generation has been terminated")
		End If
		previousX = e.AbsEndX
		previousY = e.AbsEndY
		previousZ = e.AbsEndZ

		End Sub

    ''' <summary>
    ''' This event handler is invoked for every clockwise circular interpolation move
    ''' </summary>
    ''' <param name="e">Provides data for the circular interpolation event.</param>
    Private Sub CCWInterpolation(ByVal e As CCWInterpolationEventArgs) Implements IMachineDefinition_2_0.CCWInterpolation

   		Dim moveType As MoveType = MovementType(previousX, previousY, previousZ, e.AbsEndX, e.AbsEndY, e.AbsEndZ)
		If moveType = MoveType.XY Then
			If firstCC OrElse Not Equality(previousCCX, e.AbsCentreX) OrElse Not Equality(previousCCY, e.AbsCentreY)
				NcCode.LineNumber()
				NcCode.NewString("CC X" & e.AbsCentreX.ToString(format, CultureInfo) & " Y" & e.AbsCentreY.ToString(format, CultureInfo))
				NcCode.NewLine()
				firstCC = False
			End If
			previousCCX = e.AbsCentreX
			previousCCY = e.AbsCentreY
			NcCode.LineNumber()
			NcCode.NewString("C ")
			NcCode.NewDimensionedWord("X", e.AbsEndX.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewDimensionedWord("Y", e.AbsEndY.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString("DR+ ")
			NcCode.NewString(currentCRC & " F" & e.FeedRate.ToString(0, CultureInfo) & " M")
			NcCode.NewLine()
		Else If moveType = MoveType.XYZ Then
			MessageBox.Show("Three axis CCW circular move detected", "Error", MessageBoxButtons.OK, MessageBoxIcon.Information)
			Throw New System.Exception("Nc code generation has been terminated")
		End If
		previousX = e.AbsEndX
		previousY = e.AbsEndY
		previousZ = e.AbsEndZ

    End Sub

    ''' <summary>
    ''' This event handler is invoked to drill the first hole after a tool change.
    ''' </summary>
    ''' <param name="e">Provides data for the drill first hole after toolchange event</param>
    Private Sub DrillFirstHoleAfterToolChange(ByVal e As DrillFirstHoleAfterToolChangeEventArgs) Implements IMachineDefinition_2_0.DrillFirstHoleAfterToolChange

		Me.DrillFirstHole(e, True)
		
    End Sub

    ''' <summary>
    ''' This event handler is invoked to drill the first hole in an operation that is NOT after a Tool change.
    ''' </summary>
    ''' <param name="e">Provides data for the drill first hole in operation event.</param>
    Private Sub DrillFirstHoleInOperation(ByVal e As DrillFirstHoleInOperationEventArgs) Implements IMachineDefinition_2_0.DrillFirstHoleInOperation

		Me.DrillFirstHole(New DrillFirstHoleAfterToolChangeEventArgs(e.PositionX, e.PositionY, e.DrillingType, _
																	 String.Empty, New ToolType(), 0, 0, _
																	 String.Empty, 0, String.Empty, 0 , _
																	 e.Coolant, 0, String.Empty, _
																	 e.FeedRate, e.SpindleSpeed, _
																	 e.IntialRapid, e.FeedFrom, _
																	 e.MaterialSurface, e.FinishDepth, _
																	 e.PeckAmount, e.Dwell, e.Pitch, _
																	 e.RetractToIntialRapid, e.CoolantType, _
																	 e.SpindleDirection, e.Metric), False)

    End Sub

    ''' <summary>
    ''' This event handler is invoked to drill a hole that is NOT after a Tool change or the first in the operation.
    ''' </summary>
    ''' <param name="e">Provides data for the drill hole event.</param>
    Private Sub DrillHole(ByVal e As DrillHoleEventArgs) Implements IMachineDefinition_2_0.DrillHole

        NcCode.LineNumber()
        NcCode.NewString("L ")
        NcCode.NewDimensionedWord("X", e.PositionX.ToString(format, CultureInfo), Modal.Off)
        NcCode.NewDimensionedWord("Y", e.PositionY.ToString(format, CultureInfo), Modal.Off)
        NcCode.NewString("R0 " & maxFeedrate & " M99")
        NcCode.NewLine()

    End Sub

    ''' <summary>
    ''' This event handler is invoked when a drilling cycle needs to be canelled.
    ''' </summary>
    ''' <param name="e">Provides data for the cancel drilling cycle event.</param>
	Private Sub CancelDrillingCycles(ByVal e As CancelDrillingCyclesEventArgs) Implements IMachineDefinition_2_0.CancelDrillingCycles

    End Sub

    ''' <summary>
    ''' This event handler is invoked last.
    ''' </summary>
    Private Sub ProgramEnd(ByVal e As ProgramEndEventArgs) Implements IMachineDefinition_2_0.ProgramEnd

		NcCode.LineNumber()
		NcCode.NewString("STOP M5")
	    NcCode.NewLine()

	End Sub

	''' <summary>
	''' <summary>
	''' This event handler is invoked so the start of the file can be created.
	''' </summary>
	''' <param name="e">Provides data for the program start event.</param>
	''' <remarks>Here is where you should create the end of the file, for example the % sign as used by a Fanuc.</remarks>
	Private Sub FileEnd(ByVal e As FileEndEventArgs) Implements IMachineDefinition_2_0.FileEnd

		If _configuration.OutputToolOffsetValue Then
			For index As Integer = 0 To toolOffsets.Count - 1
				For Each line As System.Text.StringBuilder In e.NcCodeLines
					line.Replace("TOOL DEF " & toolNumbers(index) & " L0 R0", _
								"TOOL DEF " & toolNumbers(index) & " L" & toolOffsets(index).ToString(format, CultureInfo) & _
								" R" & toolRadiuses(index).ToString(format, CultureInfo))
				Next
			Next index
		End If
	
	End Sub

	''' <summary>
    ''' This event handler is invoked to define the start of all Sub Routines.
    ''' </summary>
    ''' <param name="e">Provides data for the all sub routine start event.</param>
    Private Sub SubRoutineAllStart(ByVal e As SubRoutineAllStartEventArgs) Implements IMachineDefinition_2_0.SubRoutineAllStart

	End Sub

    ''' <summary>
    ''' This event handler is invoked when a Sub Routine is called.
    ''' </summary>
    ''' <param name="e">Provides data for the all sub routine start event.</param>
    Private Sub SubRoutineCall(ByVal e As SubRoutineCallEventArgs) Implements IMachineDefinition_2_0.SubRoutineCall

		NcCode.LineNumber()
        NcCode.NewString("CALL LBL" & e.SubRoutineNumber & " REP")
        NcCode.NewLine()

	End Sub

    ''' <summary>
    ''' This event handler is invoked to define the start of the Sub Routine.
    ''' </summary>
    ''' <param name="e">Provides data for the sub routine start event.</param>
    Private Sub SubRoutineStart(ByVal e As SubRoutineStartEventArgs) Implements IMachineDefinition_2_0.SubRoutineStart
        
		NcCode.LineNumber()
		NcCode.NewString("LBL" & e.SubRoutineNumber)
        NcCode.NewLine()

	End Sub

    ''' <summary>
    ''' This event handler is invoked to define the end of the Sub Routine.
    ''' </summary>
    ''' <param name="e">Provides data for the sub routine end event.</param>
    Private Sub SubRoutineEnd(ByVal e As SubRoutineEndEventArgs) Implements IMachineDefinition_2_0.SubRoutineEnd

		NcCode.LineNumber()
		NcCode.NewString("LBL0")
        NcCode.NewLine()

	End Sub

    ''' <summary>
    ''' This event handler is invoked to define the end of all Sub Routines.
    ''' </summary>
    ''' <param name="e">Provides data for the all sub routine end event.</param>
    Private Sub SubRoutineAllEnd(ByVal e As SubRoutineAllEndEventArgs) Implements IMachineDefinition_2_0.SubRoutineAllEnd

	End Sub


#End Region

#Region "Drilling Cyle Subs"

	''' <summary>
	''' Drill the first hole after a tool change or in operation
	''' </summary>
	''' <param name="e">The <see cref="DrillFirstHoleAfterToolChangeEventArgs"/> arguments</param>
	''' <param name="afterToolChange">A value indicating if the hole is the first after a tool change or in operation. <c>True</c> if first after tool change, <c>False</c> if first after operation</param>
	Private Sub DrillFirstHole(ByVal e As DrillFirstHoleAfterToolChangeEventArgs, ByVal afterToolChange As Boolean)
	
		If e.DrillingType = DrillingType.Drilling
			Cycle1(e, True, afterToolChange)
		Else If e.DrillingType = DrillingType.PeckDrill	
			Cycle1(e, False, afterToolChange)
		Else If e.DrillingType = DrillingType.Tapping	
			Cycle2(e, afterToolChange)
        End If	
        NcCode.NewLine()
		
	End Sub

	''' <summary>
	''' Output X, Y, Z position for old style cycles
	''' </summary>
	''' <param name="e">The <see cref="DrillFirstHoleAfterToolChangeEventArgs"/> arguments</param>
	''' <param name="afterToolChange">A value indicating if the hole is the first after a tool change or in operation. <c>True</c> if first after tool change, <c>False</c> if first after operation</param>
	Private Sub DrillOldCycle(ByVal e As DrillFirstHoleAfterToolChangeEventArgs, ByVal afterToolChange As Boolean)
		
		If afterToolChange Then
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("X", e.PositionX.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewDimensionedWord("Y", e.PositionY.ToString(format, CultureInfo), Modal.Off)
			If _configuration.UseM13 And e.CoolantType = Coolant.Flood And e.SpindleDirection = SpindleDirection.CW Then
				NcCode.NewString("R0 " & maxFeedrate & " M13")
			Else If _configuration.UseM13 And e.CoolantType = Coolant.Flood And e.SpindleDirection = SpindleDirection.CCW Then
				NcCode.NewString("R0 " & maxFeedrate & " M14")
			Else
				NcCode.NewString("R0 " & maxFeedrate & " M3")
			End If
			NcCode.NewLine()
			
			NcCode.LineNumber()
			NcCode.NewString("L Z" & e.IntialRapid.ToString(format, CultureInfo) & " R0 " & maxFeedrate & " ")
			If Not _configuration.UseM13 Or Not e.CoolantType = Coolant.Flood Then
	        	NcCode.NewString(e.Coolant)
			End If
			NcCode.NewLine()
			
			NcCode.LineNumber()
			NcCode.NewString("L Z" & e.FeedFrom.ToString(format, CultureInfo) & " R0 " & maxFeedrate & " M99")
        Else	
			NcCode.LineNumber()
			NcCode.NewString("L Z" & e.IntialRapid.ToString(format, CultureInfo) & " R0 " & maxFeedrate & " ")
			If Not _configuration.UseM13 Or Not e.CoolantType = Coolant.Flood Then
	        	NcCode.NewString(e.Coolant)
			End If
			NcCode.NewLine()
			
			NcCode.LineNumber()
			NcCode.NewString("L ")
			NcCode.NewDimensionedWord("X", e.PositionX.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewDimensionedWord("Y", e.PositionY.ToString(format, CultureInfo), Modal.Off)
			NcCode.NewString("R0 " & maxFeedrate & " M")
			NcCode.NewLine()

			NcCode.LineNumber()
			NcCode.NewString("L Z" & e.FeedFrom.ToString(format, CultureInfo) & " R0 " & maxFeedrate & " M99")
		End If
		NcCode.NewLine()

	End Sub

	''' <summary>
	''' Output code for Cycle 1
	''' </summary>
	''' <param name="e">The <see cref="DrillFirstHoleAfterToolChangeEventArgs"/> arguments</param>
	''' <param name="noPeck">A value indicating if the cycle is to drill to full depth and not peck. <c>True</c> for drill to full depth, <c>false</c> to peck</param>
	''' <param name="afterToolChange">A value indicating if the hole is the first after a tool change or in operation. <c>True</c> if first after tool change, <c>False</c> if first after operation</param>
	Private Sub Cycle1(ByVal e As DrillFirstHoleAfterToolChangeEventArgs, _
					   ByVal noPeck As Boolean, _
					   ByVal afterToolChange As Boolean)

	    NcCode.LineNumber()
		NcCode.NewString("CYCL DEF 1.0 PECKING")
		NcCode.NewLine()
        NcCode.LineNumber()
		NcCode.NewString("CYCL DEF 1.1 SET UP " &  e.MaterialSurface - e.FeedFrom)
		NcCode.NewLine()
        NcCode.LineNumber()
		NcCode.NewString("CYCL DEF 1.2 DEPTH " & e.FinishDepth - e.MaterialSurface)
		NcCode.NewLine()
        NcCode.LineNumber()
		If noPeck
			NcCode.NewString("CYCL DEF 1.3 PECKG " & e.FinishDepth - e.MaterialSurface)
		Else
			NcCode.NewString("CYCL DEF 1.3 PECKG " & -e.PeckAmount)
		End If
		NcCode.NewLine()
        NcCode.LineNumber()
		NcCode.NewString("CYCL DEF 1.4 DWELL " & e.Dwell)
		NcCode.NewLine()
        NcCode.LineNumber()
		NcCode.NewString("CYCL DEF 1.5 F" & e.FeedRate)
		NcCode.NewLine()
		
		Me.DrillOldCycle(e, afterToolChange)
		
	End Sub

	''' <summary>
	''' Output code for Cycle 2
	''' </summary>
	''' <param name="e">The <see cref="DrillFirstHoleAfterToolChangeEventArgs"/> arguments</param>
	''' <param name="afterToolChange">A value indicating if the hole is the first after a tool change or in operation. <c>True</c> if first after tool change, <c>False</c> if first after operation</param>
	Private Sub Cycle2(ByVal e As DrillFirstHoleAfterToolChangeEventArgs, ByVal afterToolChange As Boolean)

		NcCode.LineNumber()
		NcCode.NewString("CYCL DEF 2.0 TAPPING")
		NcCode.NewLine()
        NcCode.LineNumber()
		NcCode.NewString("CYCL DEF 2.1 SET UP " &  e.MaterialSurface - e.FeedFrom)
		NcCode.NewLine()
        NcCode.LineNumber()
		NcCode.NewString("CYCL DEF 2.2 DEPTH " & e.FinishDepth - e.MaterialSurface)
		NcCode.NewLine()
        NcCode.LineNumber()
		NcCode.NewLine()
        NcCode.LineNumber()
		NcCode.NewString("CYCL DEF 2.3 DWELL " & e.Dwell)
		NcCode.NewLine()
        NcCode.LineNumber()
		'Reduce feedrate by 5% for floating holder
		NcCode.NewString("CYCL DEF 2.4 F" & (e.Pitch * e.SpindleSpeed * 0.95).ToString(format, CultureInfo)) 
		NcCode.NewLine()
		
		Me.DrillOldCycle(e, afterToolChange)
		
	End Sub

#End Region

#Region "User Congigured Input"

    ''' <summary>
    ''' When the user clicks the 'User Configured Input' button on the NC Code tab in the Part Manager,
    ''' this sub procedure is called. Here we display the dialog box where the user can input information
	''' that can be used in the Machine Definition.
    ''' </summary>
    Private Sub UserConfiguredInput(ByVal e As UserConfiguredInputEventArgs) Implements IMachineDefinition_2_0.UserConfiguredInput

    End Sub

#End Region

#Region "Properties"

    ''' <summary>
    ''' Gets the Nc code to be output for a clockwise spindle start.
    ''' </summary>
    Private ReadOnly Property CWSpindleStart() As String Implements IMachineDefinition_2_0.CWSpindleStart
        Get
            Return "M3"
        End Get
    End Property

    ''' <summary>
    ''' Gets the Nc code to be output for a counter clockwise spindle start.
    ''' </summary>
    Private ReadOnly Property CCWSpindleStart() As String Implements IMachineDefinition_2_0.CCWSpindleStart
        Get
            Return "M4"
        End Get
    End Property

    ''' <summary>
    ''' Gets the Nc code to be output for flood coolant.
    ''' </summary>
    Private ReadOnly Property CoolantFlood() As String Implements IMachineDefinition_2_0.CoolantFlood
        Get
            Return _configuration.FloodCoolant
        End Get
    End Property

    ''' <summary>
    ''' Gets the Nc code to be output for coolant through spindle.
    ''' </summary>
    Private ReadOnly Property CoolantThroughSpindle() As String Implements IMachineDefinition_2_0.CoolantThroughSpindle
        Get
            Return _configuration.ThroughSpindleCoolant
        End Get
    End Property

    ''' <summary>
    ''' Gets the Nc code to be output for coolant through the tool.
    ''' </summary>
    Private ReadOnly Property CoolantThroughTool() As String Implements IMachineDefinition_2_0.CoolantThroughTool
        Get
            Return _configuration.ThroughToolCoolant
        End Get
    End Property

    ''' <summary>
    ''' Gets the Nc code to be output for mist coolant.
    ''' </summary>
    Private ReadOnly Property CoolantMist() As String Implements IMachineDefinition_2_0.CoolantMist
        Get
            Return _configuration.MistCoolant
        End Get
    End Property

    ''' <summary>
    ''' Gets the Nc code to be output for an air blast.
    ''' </summary>
    Private ReadOnly Property CoolantAirBlast() As String Implements IMachineDefinition_2_0.CoolantAirBlast
        Get
            Return _configuration.AirBlastCoolant
        End Get
    End Property

    ''' <summary>
    ''' Gets the Nc code to be output for coolant off.
    ''' </summary>
    Private ReadOnly Property CoolantNone() As String Implements IMachineDefinition_2_0.CoolantNone
        Get
            Return "M9"
        End Get
    End Property

    ''' <summary>
    ''' Gets the Nc code to be output for custom coolant.
    ''' </summary>
    Private ReadOnly Property CoolantCustom() As String Implements IMachineDefinition_2_0.CoolantCustom
        Get
            Return ""
        End Get
    End Property

#End Region

#Region "Modals"

    ''' <summary>
    ''' This event handler is invoked to sets up modal words
    ''' </summary>
    ''' <param name="e">Provides data for the modals event</param>
    Private Sub Modals(ByVal e As ModalEventArgs) Implements IMachineDefinition_2_0.Modals

		'Add Absolute Modal Dimentioned Words to 'AbsoluteDimensionedWords'.	
        e.AbsoluteDimensionedWords.Add(New ModalAbsoluteDimensionedWord("X"))
        e.AbsoluteDimensionedWords.Add(New ModalAbsoluteDimensionedWord("Y"))
        e.AbsoluteDimensionedWords.Add(New ModalAbsoluteDimensionedWord("Z"))
        e.AbsoluteDimensionedWords.Add(New ModalAbsoluteDimensionedWord("F"))

    End Sub

#End Region

#Region "Miscellaneous"

	''' <summary>
	''' Gets the <see cref="MoveType"/>
	''' </summary>
	''' <param name="startX">The start X value</param>
	''' <param name="startY">The start Y value</param>
	''' <param name="startZ">The start Z value</param>
	''' <param name="endX">The end X value</param>
	''' <param name="endY">The end Y value</param>
	''' <param name="endZ">The end Z value</param>
	Private Function MovementType(ByVal startX As Double, _
							      ByVal startY As Double, _
							      ByVal startZ As Double, _
							      ByVal endX As Double, _
							      ByVal endY As Double, _
							      ByVal endZ As Double) As MoveType
										
		If Equality(startX, endX) AndAlso Equality(startY, endY) AndAlso Equality(startZ, endZ)
			Return MoveType.None
		Else If Equality(startZ, endZ)
			Return MoveType.XY
		Else If  Equality(startX, endX) AndAlso Equality(startY, endY)
			Return MoveType.Z
		End If
		Return MoveType.XYZ
		
	End Function

	''' <summary>
    ''' Epsilon-based comparison for two floating point numbers.
    ''' </summary>
    ''' <param name="leftOperand">The value on the left side of the operator</param>
    ''' <param name="rightOperand">The value on the right side of the operator</param>
    ''' <returns><c>True</c> if the values are considered equal, otherwise <c>False</c></returns>
    Private Function Equality(ByVal leftOperand As Double, ByVal rightOperand As Double) As Boolean
        
		Return System.Math.Abs(leftOperand - rightOperand) <= ZERO_MAX
    
	End Function
	
	Private Enum MoveType
	
		''' <summary>
		''' No movement
		''' </summary>
		''' <returns></returns>
		None
		
		''' <summary>
		''' Three axis move in XYZ
		''' </summary>
		XYZ
			
		''' <summary>
		''' Two axis move in XY
		''' </summary>
		XY
		
		''' <summary>
		''' Move in Z axis only
		''' </summary>
		''' <returns></returns>
		Z
		
	End Enum

#End Region

End Class

#Region "Configuration"

''' <summary>
''' Class to manage the configuration of this Machine Definition.
''' </summary>
Public Class Configuration

    Private _floodCoolant As String
    Private _throughSpindleCoolant As String
    Private _throughToolCoolant As String
    Private _mistCoolant As String
    Private _airBlastCoolant As String
    Private _customCoolant As String
    Private _useM13 As Boolean
    Private _fMax As Boolean
    Private _maxFeedrate As Double
    Private _outputToolOffsetValue As Boolean


    ''' <summary>
    ''' Gets or sets the M code that turns on the Flood Coolant.
    ''' </summary>
    <System.ComponentModel.Description("The M code that turns on the Flood Coolant."), _
    System.ComponentModel.DisplayName("Flood Coolant M code"), _
    System.ComponentModel.Category("Coolant")> _
    Public Property FloodCoolant() As String
        Get
            Return _floodCoolant
        End Get
        Set(ByVal value As String)
            _floodCoolant = value
        End Set
    End Property

    ''' <summary>
    ''' Gets or sets the M code that turns on the Through Spindle Coolant.
    ''' </summary>
    <System.ComponentModel.Description("The M code that turns on the Through Spindle Coolant."), _
    System.ComponentModel.DisplayName("Through Spindle Coolant M code"), _
    System.ComponentModel.Category("Coolant")> _
    Public Property ThroughSpindleCoolant() As String
        Get
            Return _throughSpindleCoolant
        End Get
        Set(ByVal value As String)
            _throughSpindleCoolant = value
        End Set
    End Property

    ''' <summary>
    ''' Gets or sets the M code that turns on the Through Tool Coolant.
    ''' </summary>
    <System.ComponentModel.Description("The M code that turns on the Through Tool Coolant."), _
    System.ComponentModel.DisplayName("Through Tool Coolant M code"), _
    System.ComponentModel.Category("Coolant")> _
    Public Property ThroughToolCoolant() As String
        Get
            Return _throughToolCoolant
        End Get
        Set(ByVal value As String)
            _throughToolCoolant = value
        End Set
    End Property

    ''' <summary>
    ''' Gets or sets the M code that turns on the Mist Coolant.
    ''' </summary>
    <System.ComponentModel.Description("The M code that turns on the Mist Coolant."), _
    System.ComponentModel.DisplayName("Mist Coolant M code"), _
    System.ComponentModel.Category("Coolant")> _
    Public Property MistCoolant() As String
        Get
            Return _mistCoolant
        End Get
        Set(ByVal value As String)
            _mistCoolant = value
        End Set
    End Property

    ''' <summary>
    ''' Gets or sets the M code that turns on the AirBlast Coolant.
    ''' </summary>
    <System.ComponentModel.Description("The M code that turns on the AirBlast Coolant."), _
    System.ComponentModel.DisplayName("AirBlast Coolant M code"), _
    System.ComponentModel.Category("Coolant")> _
    Public Property AirBlastCoolant() As String
        Get
            Return _airBlastCoolant
        End Get
        Set(ByVal value As String)
            _airBlastCoolant = value
        End Set
    End Property

    ''' <summary>
    ''' Gets or sets the M code that turns on the Custom Coolant.
    ''' </summary>
    <System.ComponentModel.Description("The M code that turns on the Custom Coolant."), _
    System.ComponentModel.DisplayName("Custom Coolant M code"), _
    System.ComponentModel.Category("Coolant")> _
    Public Property CustomCoolant() As String
        Get
            Return _customCoolant
        End Get
        Set(ByVal value As String)
            _customCoolant = value
        End Set
    End Property

    ''' <summary>
    ''' Gets or sets a value indicating if M13/M14 should be used when Flood coolant is being used.
    ''' </summary>
    <System.ComponentModel.Description("Use M13/M14 when Food coolant is being used."), _
    System.ComponentModel.DisplayName("Use M13"), _
    System.ComponentModel.Category("Coolant")> _
    Public Property UseM13() As Boolean
        Get
            Return _useM13
        End Get
        Set(ByVal value As Boolean)
            _useM13 = value
        End Set
    End Property

    ''' <summary>
    ''' Gets or sets the maximum feedrate of the control.
    ''' </summary>
    <System.ComponentModel.Description("Maximum feedrate of the control. Used when FMAX property is set to False"), _
    System.ComponentModel.DisplayName("Maximum Feedrate"), _
    System.ComponentModel.Category("Rapids")> _
    Public Property MaxFeedrate() As Double
        Get
            Return _maxFeedrate
        End Get
        Set(ByVal value As Double)
            _maxFeedrate = value
        End Set
    End Property

    ''' <summary>
    ''' Gets or sets a value indicating if the Offset value from the Tool Manger is used in the TOOL DEF
    ''' </summary>
    <System.ComponentModel.Description("Use Offset value from Tool Manager for length in TOOL DEF"), _
    System.ComponentModel.DisplayName("Use Offet value")> _
    Public Property OutputToolOffsetValue() As Boolean
        Get
            Return _outputToolOffsetValue
        End Get
        Set(ByVal value As Boolean)
            _outputToolOffsetValue = value
        End Set
    End Property

End Class


#End Region

